/*
 *	Sherlock Indexer -- Lexicon Ordering
 *
 *	(c) 2001--2003 Martin Mares <mj@ucw.cz>
 *
 *	This module reorders the lexicon to get words which are likely
 *	to be processed together (e.g. if they differ only in accents)
 *	close to each other, thus making the reference chains close to
 *	each other as well, minimizing disk seek time.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "lib/conf.h"
#include "lib/fastbuf.h"
#include "lib/hashfunc.h"
#include "lib/math.h"
#include "lib/pools.h"
#include "charset/unicode.h"
#include "indexer/indexer.h"
#include "indexer/lexicon.h"

#include <stdlib.h>
#include <fcntl.h>

#define LH_WORDS_ONLY
#define LH_NEED_CLEANUP
#define LH_LEXORDER
#include "indexer/lexhash.h"

static struct verbum **word_array, **id_to_word;
static uns n_words, n_cplx;

static void
lex_load_words(struct fastbuf *s, uns cnt)
{
  for (uns i=0; i<cnt; i++)
    {
      byte buf[MAX_WORD_LEN+1];
      u32 origid = bgetl(s);
      ASSERT((origid&7) != WC_UNUSED && (origid&7) != WC_COMPOUND);
      u32 origcnt = bgetl(s);
      uns len = bgetc(s);
      breadb(s, buf, len);
      buf[len] = 0;
      struct verbum *v = lh_insert(buf, 0);
      if (!v)
	die("Malformed lexicon: Duplicate word <%s>", buf);
      v->id = origid;
      v->u.count = origcnt;
      ASSERT(v->id/8 <= cnt);
      word_array[i] = v;
    }
  ASSERT(lh_id == 8*cnt);
}

static void
lex_load_cplx(struct fastbuf *s, uns cnt)
{
  for (uns i=0; i<cnt; i++)
    {
      u32 origid = bgetl(s);
      ASSERT((origid&7) == WC_COMPOUND);
      u32 origcnt = bgetl(s);
      uns len = bgetc(s);
      struct verbum *v = mp_alloc_fast(lh_pool, sizeof(struct verbum) + len);
      bread(s, v->word, len);
      v->id = origid;
      v->u.count = origcnt;
      word_array[n_words+i] = v;
    }
}

static void
lex_assign_ids(struct verbum **w, uns cnt)
{
  for (uns i=0; i<cnt; i++)
    {
      struct verbum *v = *w++;
      v->id = (v->id & 7) | (8*(i+1));
    }
}

/* Maintaining equivalence classes */

/* Tarjan's Union-Find tree for word equivalence classes, tree roots have 0x80000000 + tree size */
static uns *equiv;

static uns
get_class_root(uns i)
{
  /* Get root of Tarjan's tree containing i and perform path compression along the way */
  uns j = i;
  while (!(equiv[j] & 0x80000000))
    j = equiv[j];
  while (i != j)
    {
      uns k = equiv[i];
      equiv[i] = j;
      i = k;
    }
  return i;
}

static inline void
merge_classes(uns i, uns j)
{
  i = get_class_root(i);
  j = get_class_root(j);
  if (i != j)
    {
      equiv[i] += equiv[j] & 0x7fffffff;
      equiv[j] = i;
    }
}

static struct fastbuf *f_classes;

static uns
find_class_rep(uns *cls, uns n UNUSED)
{
  /* Find the most frequent word as a representative */
  uns bestc = 0, rep = 0;
  for (uns i=0; i<n; i++)
    if (id_to_word[cls[i]]->u.count > bestc)
      {
	bestc = id_to_word[cls[i]]->u.count;
	rep = cls[i];
      }
  if (!bestc)		/* Synonymic classes can have no frequency */
    return cls[0];

#ifdef LOCAL_DEBUG
  DBG("Class: n=%d", n);
  for (uns i=0; i<n; i++)
    DBG("\t<%s> (%x) %d%s", id_to_word[cls[i]]->word, id_to_word[cls[i]]->id & 7,
	id_to_word[cls[i]]->u.count, cls[i] == rep ? " *" : "");
#endif

  if (f_classes)
    {
      bprintf(f_classes, "# Class with %d entries\n", n);
      for (uns i=0; i<n; i++)
	bprintf(f_classes, "%s%s\n", id_to_word[cls[i]]->word, cls[i] == rep ? " *" : "");
    }

  /* Check class collisions */
  struct verbum *r = id_to_word[rep];
  for (uns i=0; i<n; i++)
    if (cls[i] != rep)
      {
	struct verbum *v = id_to_word[cls[i]];
	if (((v->id ^ r->id) & 7) && v->u.count)
	  {
#ifdef CLASS_WARNINGS
	    log(L_WARN, "Words <%s> and <%s> are in the same class, categories differ (%d != %d)",
		r->word, v->word, r->id & 7, v->id & 7);
#endif
	    /* We leave the colliding pairs in the stem file as there is no easy way how to delete them */
	  }
      }

  return rep;
}

static void
find_class_reps(void)
{
  uns total = 0;
  uns nclasses = 0;
  uns maxclass = 0;
  DBG("Searching for representatives...");
  for (uns i=1; i<=n_words; i++)
    if (equiv[i] == 0x80000001)
      equiv[i] = 0x80000000;
    else if (equiv[i] > 0x80000001)
      total += equiv[i] & 0x7fffffff;
  DBG("%d words in non-trivial classes", total);

  if (fn_lex_classes)
    f_classes = bopen(index_name(fn_lex_classes), O_WRONLY|O_CREAT|O_TRUNC, indexer_fb_size);

  if (total)			/* There are non-trivial classes */
    {
      uns *cl = xmalloc(sizeof(uns) * total);
      total = 0;
      /* Prepare buckets for classes, each class starts with its root */
      for (uns i=1; i<=n_words; i++)
	if (equiv[i] > 0x80000000)
	  {
	    uns j = equiv[i] & 0x7fffffff;
	    cl[total] = i;
	    equiv[i] = 0x80000000 + total + 1;
	    total += j;
	  }
      for (uns i=1; i<=n_words; i++)
	if (equiv[i] < 0x80000000)
	  {
	    uns j = get_class_root(i);
	    ASSERT(equiv[j] > 0x80000000);
	    cl[equiv[j]++ & 0x7fffffff] = i;
	  }
      uns k = 0;
      while (k < total)
	{
	  ASSERT(equiv[cl[k]] > 0x80000000);
	  uns l = k+1;
	  while (l < total && equiv[cl[l]] < 0x80000000)
	    l++;
	  nclasses++;
	  maxclass = MAX(maxclass, l-k);
	  uns rep = find_class_rep(cl+k, l-k);
	  while (k < l)
	    equiv[cl[k++]] = rep;
	}
      xfree(cl);
    }
  /* Trivial classes are their own representatives */
  for (uns i=1; i<=n_words; i++)
    if (equiv[i] == 0x80000000)
      equiv[i] = i;
    else
      ASSERT(equiv[i] < 0x80000000);
  bclose(f_classes);
  log(L_INFO, "Found %d trivial classes + %d non-trivial classes (largest=%d)", n_words-total, nclasses, maxclass);
}

#ifdef CONFIG_LANG

#include "lang/lang.h"

static struct fastbuf *
generate_stems(void)
{
  struct fastbuf *b = bopen_tmp(indexer_fb_size);
  struct stemmer *st;
  struct mempool *mp;

  if (EMPTY_LIST(stemmer_list))
    return b;

  log(L_INFO, "Searching for stems");
  mp = mp_new(4096);
  lang_init_stemmers();
  WALK_LIST(st, stemmer_list)
    {
      DBG("Trying stemmer %d langmask %08x", st->id, st->lang_mask);
      bputl(b, st->id);
      bputl(b, st->lang_mask);
      uns npairs = 0;
      uns old_id = lh_id;
      for (uns i=0; i<n_words; i++)
	{
	  struct verbum *v = word_array[i];
	  struct stemmer_result *stem;
	  list *stems;
	  PROGRESS(i, "%s: %d words of %d, created %d new stems", st->name, i, n_words, (lh_id-old_id)/8);
	  mp_flush(mp);
	  if (stems = lang_stem(st, v->word, mp))
	    WALK_LIST(stem, *stems)
	      {
		struct verbum *w = lh_lookup_utf8(stem->stem);
		if (v == w)
		  ;
		else if (w->id < 8)
		  log(L_WARN, "Word <%s> has too long stem <%s>", v->word, stem->stem);
		else
		  {
		    bputl(b, w->id);
		    bputl(b, v->id);
		    npairs++;
		  }
	      }
	}
      bputl(b, ~0U);
      log(L_INFO, "Stemmer %s (langmask %x): Processed %d words, created %d new stems",
	  st->name, st->lang_mask, npairs, (lh_id-old_id)/8);
    }
  bflush(b);
  mp_delete(mp);
  return b;
}

static void
generate_syn_classes(struct fastbuf *b)
{
  struct syndict *s;
  struct mempool *mp;
  byte **cls;

  if (EMPTY_LIST(syndict_list))
    return;

  mp = mp_new(4096);
  log(L_INFO, "Constructing synonymic classes");
  struct fastbuf *c = bopen(b->name, O_RDONLY, indexer_fb_size);
  bsetpos(c, btell(b));
  WALK_LIST(s, syndict_list)
    {
      uns nclasses = 0;
      uns nwords = 0;
      uns old_id = lh_id;

      /* Construct direct map */
      syndict_open(s);
      bputl(b, 0x80000000);
      bputl(b, s->lang_mask);
      while (cls = syndict_read_entry(s, mp))
	{
	  PROGRESS(nclasses, "%s: %d classes", s->name, nclasses);
	  nclasses++;
	  if (!cls[0] || !cls[1])
	    continue;
	  struct verbum *w = NULL;
	  for (uns i=0; cls[i]; i++)
	    {
	      struct verbum *v = lh_lookup_utf8(cls[i]);
	      if (v->id < 8)
		log(L_WARN, "Word <%s> is too long", cls[i]);
	      else if (w && ((v->id ^ w->id) & 7))
#ifdef CLASS_WARNINGS
		log(L_WARN, "Synonyms <%s> and <%s> have different classes (%d != %d)", v->word, w->word, v->id&7, w->id&7);
#else
	        ;
#endif
	      else
		{
		  bputl(b, nclasses);
		  bputl(b, v->id);
		  w = v;
		  nwords++;
		}
	    }
	}
      syndict_close(s);
      bputl(b, ~0U);
      bflush(b);

      /* Construct reverse map */
      bgetl(c);
      bgetl(c);
      bputl(b, 0x80000001);
      bputl(b, s->lang_mask);
      u32 x, y;
      while ((x = bgetl(c)) != ~0U)
	{
	  y = bgetl(c);
	  bputl(b, y);
	  bputl(b, x);
	}
      bputl(b, ~0U);

      log(L_INFO, "Synonymic dictionary %s (langmask %x): Processed %d classes (%d words, %d new)",
	  s->name, s->lang_mask, nclasses, nwords, (lh_id-old_id)/8);
    }
  bflush(b);
  bclose(c);
  mp_delete(mp);
}

static void
stem_equivs(struct fastbuf *s)
{
  u32 x, y;

  bsetpos(s, 0);
  while ((x = bgetl(s)) < 0x80000000)
    {
      bgetl(s);
      while ((x = bgetl(s)) != ~0U)
	{
	  y = bgetl(s);
	  merge_classes(x/8, y/8);
	}
    }
}

static void
renumber_stems(struct fastbuf *s)
{
  struct fastbuf *d = bopen(index_name(fn_stems_ordered), O_WRONLY|O_CREAT|O_TRUNC, indexer_fb_size);
  u32 stid, langmask, x, y;

  bsetpos(s, 0);
  while ((stid = bgetl(s)) != ~0U)
    {
      langmask = bgetl(s);
      bputl(d, stid);
      bputl(d, langmask);
      while ((x = bgetl(s)) != ~0U)
	{
	  y = bgetl(s);
	  if (stid != 0x80000000)
	    x = id_to_word[x/8]->id;
	  bputl(d, x);
	  if (stid != 0x80000001)
	    y = id_to_word[y/8]->id;
	  bputl(d, y);
	}
      bputl(d, ~0U);
    }
  bclose(s);
  bclose(d);
}

#endif

/* Building equivalence classes of words differing by accents -- sort in accentless order first */

static inline int
accent_lt(struct verbum *x, struct verbum *y)
{
  byte *xx = x->word;
  byte *yy = y->word;
  uns xu, yu, xuu, yuu;
  int pass2 = 0;

  while (1)
    {
      GET_UTF8(xx, xu);
      GET_UTF8(yy, yu);
      if (!xu || !yu)
	break;
      xuu = Uunaccent(xu);
      yuu = Uunaccent(yu);
      if (xuu < yuu)
	return 1;
      if (xuu > yuu)
	return 0;
      if (!pass2)
	{
	  if (xu < yu)
	    pass2 = -1;
	  else if (xu > yu)
	    pass2 = 1;
	}
    }
  if (yu)
    return 1;
  if (xu)
    return 0;
  if (pass2 < 0)
    return 1;
  else if (pass2 > 0)
    return 0;
  ASSERT(x == y);
  return 0;
}

static inline int
accentless_same(struct verbum *x, struct verbum *y)
{
  byte *xx = x->word;
  byte *yy = y->word;
  uns xu, yu;

  while (1)
    {
      GET_UTF8(xx, xu);
      GET_UTF8(yy, yu);
      if (!xu || !yu)
	return (!xu && !yu);
      if (Uunaccent(xu) != Uunaccent(yu))
	return 0;
    }
}

#define ASORT_PREFIX(id) accent_##id
#define ASORT_KEY_TYPE struct verbum *
#define ASORT_ELT(i) word_array[i]
#define ASORT_LT(x,y) accent_lt(x,y)
#include "lib/arraysort.h"

static void
accent_equivs(void)
{
  accent_sort(n_words);
  for (uns i=1; i<n_words; i++)
    if (accentless_same(word_array[i-1], word_array[i]))
      merge_classes(word_array[i-1]->id/8, word_array[i]->id/8);
}

/* Sort words first according to classes, then in accentless order and finally in accented order */

static inline int
word_lt(struct verbum *x, struct verbum *y)
{
  uns xclass = equiv[x->id/8];
  uns yclass = equiv[x->id/8];
  if (xclass != yclass)
    {
      x = id_to_word[xclass];
      y = id_to_word[yclass];
    }
  return accent_lt(x, y);
}

#define ASORT_PREFIX(id) word_##id
#define ASORT_KEY_TYPE struct verbum *
#define ASORT_ELT(i) word_array[i]
#define ASORT_LT(x,y) word_lt(x,y)
#include "lib/arraysort.h"

/* Sorting of the whole lexicon */

static inline uns
lex_lt_get_root(struct verbum *x, uns *xc)
{
  if ((x->id & 7) != WC_COMPOUND)
    {
      *xc = 0;
      return x->id;
    }

  u32 *c = (u32*) x->word;
  *xc = ch_wlen(c);
  if ((c[0] & 7) == WC_NORMAL)
    return id_to_word[c[0]/8] -> id;
  else
    return id_to_word[c[*xc-1]/8] -> id;
}

static int
lex_lt(struct verbum *x, struct verbum *y)
{
  if (x == y)
    return 0;

  uns xc, yc;
  uns xr = lex_lt_get_root(x, &xc);
  uns yr = lex_lt_get_root(y, &yc);

  if (xr < yr)
    return 1;
  if (xr > yr)
    return 0;
  if (xc < yc)
    return 1;
  if (xc > yc)
    return 0;
  ASSERT(xc);
  for (uns i=0; i<xc; i++)
    {
      uns a = id_to_word[((u32*)x->word)[i] / 8] -> id;
      uns b = id_to_word[((u32*)y->word)[i] / 8] -> id;
      if (a < b)
	return 1;
      if (a > b)
	return 0;
    }
  ASSERT(0);
}

#define ASORT_PREFIX(id) lex_##id
#define ASORT_KEY_TYPE struct verbum *
#define ASORT_ELT(i) word_array[i]
#define ASORT_LT(x,y) lex_lt(x,y)
#include "lib/arraysort.h"

/* Squeezing of word frequencies to range [0,255] */

static void
lex_squeeze_freqs(void)
{
  uns maxf = 0;
  for (uns i=0; i<n_words+n_cplx; i++)
    maxf = MAX(maxf, word_array[i]->u.count);
  double maxflog = log2(maxf+2);
  for (uns i=0; i<n_words+n_cplx; i++)
    {
      uns cnt = word_array[i]->u.count;
      if (cnt)
	{
	  int logf = 1 + (int)(.5 + 254 * log2(cnt) / maxflog);
	  word_array[i]->u.count = CLAMP(logf, 1, 255);
	}
    }
}

int
main(int argc, char **argv)
{
  log_init(argv[0]);
  setproctitle_init(argc, argv);
  if (cf_getopt(argc, argv, CF_SHORT_OPTS, CF_NO_LONG_OPTS, NULL) >= 0 ||
      optind < argc)
  {
    fputs("This program supports only the following command-line arguments:\n" CF_USAGE, stderr);
    exit(1);
  }

  /* Load words from the lexicon to memory */
  struct fastbuf *s = bopen(index_name(fn_lex_raw), O_RDONLY, indexer_fb_size);
  n_words = bgetl(s);
  n_cplx = bgetl(s);
  log(L_INFO, "Loading lexicon: %d words", n_words);
  word_array = xmalloc(sizeof(struct verbum *) * n_words);
  lh_init();
  lex_load_words(s, n_words);

#ifdef CONFIG_LANG
  /* Generate stem list for all stemmers on all known words, new words can arise */
  struct fastbuf *stems_tmp = generate_stems();
  /* Generate synonymic classes */
  generate_syn_classes(stems_tmp);
#endif

  /* Generate array of all words indexed by original word ID */
  xfree(word_array);
  n_words = lh_id/8;
  DBG("Gathering IDs for %d words", n_words);
  id_to_word = xmalloc(sizeof(struct verbum *) * (n_words+1));
  uns cnt = 0;
  for (uns i=0; i<lh_hash_size; i++)
    for (struct verbum *l=lh_hash_table[i]; l; l=l->next)
      {
	id_to_word[l->id/8] = l;
	cnt++;
      }
  ASSERT(cnt == n_words);
  lh_cleanup();

  /* Allocate word array for both words and complexes */
  word_array = xmalloc(sizeof(uns) * (n_words + n_cplx));
  memcpy(word_array, id_to_word+1, sizeof(uns) * n_words);

  /* Now we know all the words, so let's group them to equivalence classes and find their representatives */
  log(L_INFO, "Constructing equivalence classes for %d words", n_words);
  equiv = xmalloc(sizeof(uns) * (n_words+1));
  for (uns i=1; i<=n_words; i++)
    equiv[i] = 0x80000001;
  log(L_INFO, "Merging classes according to accents");
  accent_equivs();
#ifdef CONFIG_LANG
  log(L_INFO, "Merging classes according to stems");
  stem_equivs(stems_tmp);
#endif
  find_class_reps();

  /* Sort the word array and assign new IDs to words */
  log(L_INFO, "Sorting words");
  word_sort(n_words);
  lex_assign_ids(word_array, n_words);

  /* Load complexes to memory */
  log(L_INFO, "Loading lexicon: %d complexes", n_cplx);
  lex_load_cplx(s, n_cplx);
  bclose(s);

  /* Sort both words and complexes and assign new IDs, this time to both words and complexes */
  log(L_INFO, "Sorting whole lexicon");
  lex_sort(n_words+n_cplx);
  lex_assign_ids(word_array, n_words+n_cplx);
  lex_squeeze_freqs();

  /* Write out everything we know, renumbering word ID occurences inside complexes as we go */
  log(L_INFO, "Writing final lexicon");
  struct fastbuf *d = bopen(index_name(fn_lex_ordered), O_WRONLY|O_CREAT|O_TRUNC, indexer_fb_size);
  bputl(d, n_words);
  bputl(d, n_cplx);
  for (uns i=0; i<n_words+n_cplx; i++)
    {
      struct verbum *v = word_array[i];
      bputl(d, v->id);
      bputl(d, v->u.count);
      if ((v->id & 7) == WC_COMPOUND)
	{
	  uns c = ch_wlen((u32*)v->word);
	  bputc(d, 4*c);
	  for (uns x=0; x<c; x++)
	    {
	      u32 oldid = ((u32*)v->word)[x];
	      ASSERT(oldid/8 <= n_words);
	      bputl(d, id_to_word[oldid/8]->id);
	    }
	}
      else
	{
	  uns c = str_len(v->word);
	  bputc(d, c);
	  bwrite(d, v->word, c);
	}
    }
  bclose(d);

#ifdef CONFIG_LANG
  /* Renumber the stem list */
  log(L_INFO, "Writing stems");
  renumber_stems(stems_tmp);
#endif

  return 0;
}
