/*
 *	Sherlock Gatherer -- ObjDump
 *
 *	(c) 2001--2003 Robert Spalek <robert@ucw.cz>
 *	(c) 2003 Martin Mares <mj@ucw.cz>
 */

#include "lib/lib.h"
#include "lib/conf.h"
#include "lib/fastbuf.h"
#include "lib/bucket.h"
#include "lib/index.h"
#include "lang/detect.h"
#include "charset/unicode.h"
#include "charset/charconv.h"
#include "utils/dumpconfig.h"

#include <stdlib.h>
#include <string.h>

static int term_charset_id;
static struct conv_context conv_utf8;
static int do_lang_detect;
static int verbose;

/* Maximal input line length.  */
#define	BUFSIZE	1024

static char *options = CF_SHORT_OPTS "rlv";

static char *help = "\
Usage: objdump <options>\n\
\n\
Options:\n"
CF_USAGE
"-r\t\tRaw dump (don't convert character sets)\n\
-l\t\tLanguage detection\n\
-v\t\tBe verbose\n\
";

static void NONRET
usage(void)
{
	fputs(help, stderr);
	exit(1);
}

static byte outtext[BUFSIZE];			/* formatted output text */
static byte *outend = outtext;
static uns out_space;
static uns this_attr;

static void
flush_outtext(void)
{
	if (outend == outtext)
		return;
	printf("%c%s\n", this_attr, outtext);
	outtext[0] = 0;
	outend = outtext;
}

static void
add_word(byte *b, byte *e)
{
	if (outend-outtext + 1 + e-b >= (int)(line_len - 1))
		flush_outtext();
	if (out_space && outend != outtext)
		*outend++ = ' ';
	strncpy(outend, b, e-b);
	outend[e-b] = 0;
	outend += e-b;
	out_space = 0;
}

static char *wt_names[16] = { WORD_TYPE_USER_NAMES };
static char *mt_names[16] = { META_TYPE_USER_NAMES };

static void
objdump_line(byte *line)
{
	if (!line[0])
		return;
	if (line[0] != 'X')
		flush_outtext();
	if (line[0] == 'Z')
	{
		struct conv_context *cc = &conv_utf8;
		byte recoded[BUFSIZE];
		int flags;
		putchar('Z');
		cc->source = line + 1;
		cc->source_end = line + strlen(line);
		cc->dest = cc->dest_start = recoded;
		cc->dest_end = recoded+BUFSIZE;
		do
		{
			flags = conv_run(cc);
			if (flags & (CONV_SOURCE_END | CONV_DEST_END))
			{
				fwrite(cc->dest_start, 1, cc->dest - cc->dest_start, stdout);
				cc->dest = recoded;
			}
		}
		while (! (flags & CONV_SOURCE_END));
		putchar('\n');
	}
	else if (line[0] == 'X' || line[0] == 'M')
	{
		int is_meta = (line[0] == 'M');
		char **type_names = is_meta ? mt_names : wt_names;
		struct conv_context *cc = &conv_utf8;
		byte recoded[BUFSIZE], *end;
		int flags;
#ifdef CONFIG_LANG
		if (line[0] == 'X' && do_lang_detect)
			lang_detect_add_string(line + 1);
#endif
		cc->dest = cc->dest_start = recoded;
		cc->dest_end = recoded+BUFSIZE;
		cc->source = line + 1;
		end = line + strlen(line);
		out_space = 1;
		this_attr = line[0];
		while (cc->source < end)
		{
			byte *c;
			uns tag;
			int len;
			c = (byte*)cc->source;
			do
			{
				cc->source_end = c;
				GET_TAGGED_CHAR(c, tag);
			}
			while (tag && tag != ' ' && tag < 0x80000000);
			do
			{
				flags = conv_run(cc);
				if (flags & (CONV_SOURCE_END | CONV_DEST_END))
				{
					if (cc->dest != cc->dest_start)
						add_word(cc->dest_start, cc->dest);
					cc->dest = recoded;
				}
			}
			while (! (flags & CONV_SOURCE_END));
			cc->source = c;
			if (tag < 0x80000000)	/* space, end of string */
			{
				out_space = 1;
				continue;
			}
			if (tag < 0x80010000)
			{
				if (tag & 0x10)
				{
					out_space = 1;
					len = sprintf(recoded, "<break>");
					add_word(recoded, recoded+len);
				}
				out_space = 1;
				len = sprintf(recoded, "<%s>", type_names[tag & 0x0f] ? : "???");
				add_word(recoded, recoded+len);
				out_space = 1;
			}
			else if (tag < 0x80020000)
			{
				len = sprintf(recoded, "<ref %d>", tag & 0x03ff);
				add_word(recoded, recoded+len);
			}
			else
			{
				len = sprintf(recoded, "</ref>");
				add_word(recoded, recoded+len);
			}
		}
		if (is_meta)
			flush_outtext();
	}
	else
		printf("%s\n", line);
}

static void
objdump_done(void)
{
	flush_outtext();
	printf("\n");
#ifdef CONFIG_LANG
	if (do_lang_detect)
	{
		struct lang_detect_results *res = lang_detect_compute();
		if (res->lang1 < 0)
			printf("The language cannot be detected\n\n");
		else
		{
			struct lang_detect_lang_flag *l1, *l2;
			l1 = lang_detect_lang_flags + res->lang1;
			l2 = lang_detect_lang_flags + res->lang2;
			printf("Language detection:\n");
			printf("1st match: #%d (%s%s, %d), variance %d\n",
				res->lang1, lang_code_to_name(l1->id), l1->is_accented ? "" : "-CA",
				l1->id, res->variances[res->lang1]);
			printf("2nd match: #%d (%s%s, %d), variance %d\n",
				res->lang2, lang_code_to_name(l2->id), l2->is_accented ? "" : "-CA",
				l2->id, res->variances[res->lang2]);
			printf("Ratio of variances %d, minimal ratio %d\n", res->ratio, res->min_ratio);
			printf("Found %d occurences of %d sequences\n\n", res->total_occur, res->nonzero_seq);
			if (verbose > 0)
			{
				printf("Variances of languages:\n");
				for (uns i=0; i<lang_detect_nr_langs; i++)
				{
					struct lang_detect_lang_flag *l = lang_detect_lang_flags + i;
					printf("#%d (%s%s, %d), variance %d (ratio %d)\n",
						i, lang_code_to_name(l->id), l->is_accented ? "" : "-CA",
						l->id, res->variances[i],
						1000 * res->variances[i] / res->variances[res->lang1]);
				}
				printf("\n");
			}
			if (verbose > 1)
			{
				printf("Occurences of sequences:\n");
				for (uns i=0; i<res->nonzero_seq; i++)
				{
					if (verbose == 2 && i >= lang_detect_lang_flags[ res->lang1 ].nr_seq)
					{
						printf("Cut here\n");
						break;
					}
					struct lang_detect_sequence *seq = lang_detect_sequences[ res->sf[i].id ];
					printf("%d. %s, %d occurences, expected at [%d, %d]\n",
						i+1, seq->text, res->sf[i].occur,
						seq->order[res->lang1], seq->order[res->lang2]);
				}
				printf("\n");
			}
		}
		lang_detect_start();
	}
#endif
}

static void
dump_stdin(void)
{
	struct fastbuf *b;
	byte buf[BUFSIZE];
	b = bfdopen_shared(0, 4096);
	bgets(b, buf, BUFSIZE);
	if (strncmp(buf, "###", 3))		/* a single object */
	{
		do
		{
			objdump_line(buf);
		}
		while (bgets(b, buf, BUFSIZE));
		objdump_done();
	}
	else					/* multiple objects divided by ### lines */
	{
		do
		{
			oid_t oid;
			sscanf(buf+4, "%x", &oid);
			printf("O%x\n", oid);
			while(1)
			{
				if (!bgets(b, buf, BUFSIZE))
				{
					buf[0] = 0;
					break;
				}
				if (!strncmp(buf, "###", 3))
					break;
				objdump_line(buf);
			}
			objdump_done();
		}
		while (buf[0]);
	}
	bclose(b);
}

int
main(int argc, char **argv)
{
	int opt;

	log_init(argv[0]);
	while ((opt = cf_getopt(argc, argv, options, CF_NO_LONG_OPTS, NULL)) >= 0)
		switch (opt)
		{
			case 'r':
				term_charset_id = CONV_CHARSET_UTF8;
				break;
			case 'l':
				do_lang_detect++;
				break;
			case 'v':
				verbose++;
				break;
			default:
				usage();
		}
	if (optind < argc)
		usage();
	if (!term_charset_id)
	{
		term_charset_id = find_charset_by_name(terminal_charset);
		if (term_charset_id < 0)
			die("Unknown terminal charset %s", terminal_charset);
	}
	conv_init(&conv_utf8);
	conv_set_charset(&conv_utf8, CONV_CHARSET_UTF8, term_charset_id);
#ifdef CONFIG_LANG
	if (do_lang_detect)
	{
		lang_detect_build_automaton();
		lang_detect_start();
	}
#endif
	dump_stdin();
	return 0;
}
